#!/usr/bin/perl

$| = 1;

# ddrlogand.pl   ("DDR LOGfile logical .AND.")
#
# This is a Perl program which reads two GNU ddrescue 1.x "log files" (really,
# sector list files), and does a logical AND on them.  In other words, in the
# resulting output file, a sector is only shown as "recovered" if it was
# recovered in BOTH of the two input files.
#
# The purpose is to deal with "infant mortality" in a replacement disk drive.
# I was trying to recover a failing 160 GB Maxtor drive onto a brand new
# 200 GB Seagate drive, and had successfully copied about 137 GB with ddrescue,
# when ddrescue failed with a "write error."  So I then got another new
# Seagate drive, and used ddrescue to copy most of the old new drive onto the
# new new drive.  But it didn't get everything.  So I needed a tool to combine
# the two ddrescue log files, to reflect the recovered data that was actually
# on the new new drive, before I could resume recovering from the old Maxtor.
#
# Note that the operation is not quite symetrical:
#
# Old dying drive  -->  New dying drive  -->  New new drive  -->
#               log_input1            log_input2              result
#               ----------            ----------              ------
#  if sector is     -                  anything                  -
#  if sector is     /                  anything                  /
#  if sector is     +                     +                      +
#  if sector is     +                     -                      /
#  if sector is     +                     /                      /


# BTW, here's a hint.  Open a console and do:
#   tail -s 3 -f /var/log/messages
# to watch the messages logged during a ddrescue attempt.


# This program is UNCOPYRIGHTED.  It was written by:
#
#    Dave Burton
#    Burton Systems Software
#    Tel: +1 919-481-0149
#    Email: dave at burtonsys.com
#


#  =>keyflag<=-- "%v, %d"
$ver_ordinal_and_date = "8, 13-Sep-06";
# the TLIB Version Control version number and last-modified date


$verbose = 0;  # will be changed if '-v' is specified


$perlver = "3 or earlier";
if ($] =~ /\$\$Revision\:\s*([0-9.]+)\s/) {
   $perlver = $1;  # probably 4.something
} elsif ($] =~ /([0-9][0-9.]*)/) {
   $perlver = $1;  # probably 5.something or 6.something
}


@sv_ARGV = @ARGV;  # save raw parameter list, for display to the user

# display command-line parameters (just once)
$args_shown = 0;
sub showargs {
   if (!$args_shown) {
      print join( " ", "perl $0", @sv_ARGV ) . "\n";  # echo command-line arguments
      $args_shown = 1;
   }
}


# Handle input parameters and open all the files:

while (($#ARGV >= 0) && ($ARGV[0] =~ /^-v/)) {
   shift @ARGV;
   $verbose++;
}
($log_output, $log_input1, $log_input2) = @ARGV;

if ($verbose) {
   &showargs;
   print "You are using Perl version $perlver\n";
}

if (!defined $log_input1) {
   die "ddrlogand.pl   (\"DDRescue LOGfile logical .AND.\")   v.$ver_ordinal_and_date\n" .
       "\n" .
       "Usage:\n" .
       "   perl -w $0 outputfile inputfile1 inputfile2\n" .
       "\n" .
       "or, for verbose output:\n" .
       "   perl -w $0 -v outputfile inputfile1 inputfile2\n" .
       "\n" .
       "*** Note the odd parameter order (output file first)! ***\n" .
       "\n" .
       "Also, note that '-v' can be repeated for very verbose output.\n";
}


&showargs;

open( LOG_INPUT1, "<$log_input1") || die "ERROR: Could not read $log_input1, $!\n";
open( LOG_INPUT2, "<$log_input2") || die "ERROR: Could not read $log_input2, $!\n";
open( LOG_OUTPUT, ">$log_output" ) || die "ERROR: Could not create $log_output, $!\n";
binmode LOG_OUTPUT;  # write output in binary mode, to force LF-only line termination even under DOS/Windows



######################################################################
# Here's a sample of what a ddrescue log file looks like:
#
#    # Rescue Logfile. Created by GNU ddrescue version 1.1
#    #      pos        size  status
#    0x00000000  0x0004BE00  +
#    0x0004BE00  0x00000200  -
#    0x0004C000  0x33078E00  +
#    0x330C4E00  0x00001000  -
#    0x330C5E00  0x8ACDF200  +
#
######################################################################


# input is an integer represented as a string (probably hexadecimal preceeded by '0x'),
# result is the integer divided by 512.
sub divby512 {
   local($dividend) = @_;
   # print "dividend=$dividend\n";
   local($result) = 0;
   if ($dividend =~ /^0x/) {
      # print "tricky way!\n";
      # tricky divide by 512: drop the last 2 hexidecimal digits, and divide by 2 (which is more robust for very large integers)
      $result = substr($dividend,0,length($dividend)-2);
      $result = int(eval($result) / 2);  # gotta use eval() because it starts with 0x
   } else {
      # print "conventional way,\n";
      $result = int((0+$dividend) / 512);
   }
   # print "result=$result\n";
   return $result;
}


# Internal representation of the sector list is a bit tricky.
# The problem is that there are soooo many sectors.  A 300 GB
# disk drive has 600 million sectors.  Even at one character
# per sector, that's too big to represent in memory as a simple
# array or string.  So we have to represent it like it is in
# the sector list file, as a list of ranges.  But that means
# that we can't use a simple array or hash lookup to check the
# status of a particular sector; we'll have to call a function
# to do a search.



# First, read log_input1 (the older of the two ddrescue sector lists)
$numranges = 0;
$lastline = '';
$highest_from_inp1 = 0;
while (<LOG_INPUT1>) {
   if (($_ =~ /\d/) && ($_ !~ /^\#/)) {
      $_ =~ s/\s*$//;  # strip trailing blanks, tabs, cr, lf, etc.
      # Parse a ddrescue log file line:
      @tmp = split( /\s+/, $_ );
      # for ($i=0; $i<=$#tmp; $i++) { print "dbg: tmp[$i]='" . $tmp[$i] . "'\n"; }
      ($pos_byt, $len_byt, $status) = @tmp;
      if (($#tmp != 2) || ($pos_byt !~ /\d/) || ($len_byt !~ /\d/)) {
         print "ERROR: unrecognized line in '$log_input1':\n$_\n";
         next;
      }
      if (('-' ne $status) && ('+' ne $status)) {
         # ignore all but '-' (error) regions and '+' (successful) regions
         $lastline = $_;
         next;
      }
      $lastline = '';
      if ($len_byt eq '-1') {
         print "ERROR: length=-1 for status = '$status' in '$log_input1':\n$_\n";
         next;
      }

      # Now convert byte offsets and counts to LBA sector number & sector count:
      $pos_sect = &divby512($pos_byt);
      $len_sect = &divby512($len_byt);

      # print "dbg: Bad sect no. = $pos_sect, count = $len_sect, '$status'\n";
      # for ($i=0; $i<$len_sect; $i++) {
      #   $originally_bad_sectors{$pos_sect} = 1;
      #   $num_originally_bad_sectors++;
      #   $pos_sect++;
      # }

      $offsets[$numranges] = $pos_sect;
      $lengths[$numranges] = $len_sect;
      $stati[$numranges] = $status;  # '-' or '+'
      $numranges++;
      $highest_from_inp1 = $pos_sect + $len_sect;
   }
}


$count_rescued1 = $count_failed1 = 0;
for ($i=0; $i<$numranges; $i++) {
   $status = $stati[$i];
   $ofs = $offsets[$i];
   $len = $lengths[$i];
   if ('+' eq $status) {
      $count_rescued1 += $len;
   } elsif ('-' eq $status) {
      $count_failed1 += $len;
   }
}


# Now read the newer of the two lists:
$numranges2 = 0;
$lastline2 = '';
while (<LOG_INPUT2>) {
   if (($_ =~ /\d/) && ($_ !~ /^\#/)) {
      $_ =~ s/\s*$//;  # strip trailing blanks, tabs, cr, lf, etc.
      # Parse a ddrescue log file line:
      @tmp = split( /\s+/, $_ );
      ($pos_byt, $len_byt, $status) = @tmp;
      if (($#tmp != 2) || ($pos_byt !~ /\d/) || ($len_byt !~ /\d/)) {
         print "ERROR: unrecognized line in '$log_input2':\n$_\n";
         next;
      }
      if ('+' eq $status) {
         # ignore all but '-' (error) regions and '/' (uncopied) regions
         $lastline2 = $_;
         next;
      }
      $lastline2 = '';

      # Now convert byte offsets and counts to LBA sector number & sector count:
      $pos_sect = &divby512($pos_byt);
      if ($len_byt eq '-1') {
         $len_sect = $highest_from_inp1;
      } else {
         $len_sect = &divby512($len_byt);
      }
      $offsets2[$numranges2] = $pos_sect;
      $lengths2[$numranges2] = $len_sect;
      $numranges2++;
   }
}


# Format a ddrescue logfile line.  Input parameters are in sectors (512 bytes),
# but the result is displayed in bytes.
sub displayrange {
   local( $ofs, $len, $status ) = @_;
   local( $result ) = sprintf( "0x%06X00  0x%06X00  %s", (2*$ofs), (2*$len), $status );
   return $result;
}


# Now we get to merge the two sector lists.
# Step through the older sector list:
for ($i=0; $i<$numranges; $i++) {
   # the only regions we care about are the '+' regions.
   if ('+' eq $stati[$i]) {
      $status = '+';
      $ofs = $offsets[$i];
      $len = $lengths[$i];
      # was any part of this region not successfully copied
      # from the old new drive to the new new drive?
      for ($j=0; $j<$numranges2; $j++) {
         $ofs2 = $offsets2[$j];
         $len2 = $lengths2[$j];
         if (($ofs2+$len2) <= $ofs) {
            next;
         }
         if ($ofs2 >= ($ofs+$len)) {
            if ($verbose >= 2) {
               print "dbg: i=$i, no overlap of:\n" .
                     "  Old: " . &displayrange($ofs,$len,$status) . "\n";
               if ($j > 0) {
                  print "  new: " . &displayrange($offsets2[$j-1],$lengths2[$j-1],'-') . "\n";
               }
               print "  New: " . &displayrange($ofs2,$len2,'-') . "\n";
            }
            last;  # since the regions are in order, we're done
         }
         if ($verbose) {
            print "dbg: i=$i, j=$j, found an overlap.\n" .
                  "  Old: " . &displayrange($ofs,$len,$status) . "\n" .
                  "  New: " . &displayrange($ofs2,$len2,'-') . "\n";
         }
         # They overlap!  Drat, some of the sectors that were successfully
         # copied the first time, were lost when the second drive failed.
         # There are four cases:
         # 1) The failure invalidates the entire good region
         # 2) The failure invalidates the beginning of the good region
         # 3) The failure invalidates the end of the good region
         # 4) The failure invalidates the middle, but not the beginning or end
         if (($ofs2 <= $ofs) && (($ofs2+$len2) >= ($ofs+$len))) {
            # Case #1: the failure invalidates the entire good region
            $stati[$i] = $status = '/';
            if ($verbose) {
               print "  Entire old range invalidated.\n";
            }
         } elsif ($ofs2 <= $ofs) {
            # Case #2: the failure invalidates the beginning of the good region
            $offsets[$i] = $ofs = $ofs2+$len2;
            if ($verbose) {
               print "  Beginning of old range invalidated.\n";
            }
         } elsif (($ofs2+$len2) >= ($ofs+$len)) {
            # Case #3: the failure invalidates the end of the good region
            $lengths[$i] = $len = $ofs2 - $ofs;
            if ($verbose) {
               print "  End of old range invalidated.\n";
            }
         } else {
            # Case #4 is the hardest, because we need to split the good region
            splice( @offsets, $i+1, 0, $ofs2+$len2 );
            splice( @lengths, $i+1, 0, ($ofs+$len)-($ofs2+$len2) );
            splice( @stati, $i+1, 0, '+' );
            $lengths[$i] = $ofs2 - $ofs;
            $numranges++;
            if ($verbose) {
               print "  Middle of old range invalidated.  Split old range.\n";
               print "    i=$i, len = $len -> " . $lengths[$i] . "\n";
               print "  Beg: " . &displayrange( $offsets[$i], $lengths[$i], '+' ) . "\n" .
                     "  End: " . &displayrange( $offsets[$i+1], $lengths[$i+1], '+' ) . "\n";
            }
            $len = $lengths[$i];
         }
      }
   }
}




$count_rescued = $count_failed = 0;
for ($i=0; $i<$numranges; $i++) {
   $status = $stati[$i];
   $ofs = $offsets[$i];
   $len = $lengths[$i];
   if ('+' eq $status) {
      $count_rescued += $len;
   } elsif ('-' eq $status) {
      $count_failed += $len;
   }
}
print "$count_rescued1 sectors were originally rescued.\n";
print $count_rescued1-$count_rescued . " of them were lost when the new drive failed.\n";
print "$count_rescued remaining rescued sectors.\n";
print "$count_failed unreadable sectors.\n";


# Now spit out the list:

print LOG_OUTPUT "# Rescue Logfile. Created by GNU ddrescue version 1.1-burtonhack3\n" .
      "#      pos        size  status\n";
$prev_ofs_plus_len = 0;
for ($i = 0; $i < $numranges; $i++) {
   $ofs = $offsets[$i];
   $len = $lengths[$i];
   $status = $stati[$i];
   if ('/' eq $status) {
      next;
   }
   if ($ofs < $prev_ofs_plus_len) {
      print STDERR "ERROR: overlapping regions:\n  " . &displayrange($ofs,$len,$status) . "\n";;
   } elsif ($ofs > $prev_ofs_plus_len) {
      $gaplen = $ofs - $prev_ofs_plus_len;
      print LOG_OUTPUT &displayrange($prev_ofs_plus_len,$gaplen,'/') . "\n";
   }
   print LOG_OUTPUT &displayrange($ofs,$len,$status) . "\n";
   $prev_ofs_plus_len = $ofs + $len;
}

if ($lastline ne '') {
   ($pos_byt) = split( /\s+/, $lastline );
   $ofs = &divby512($pos_byt);
   if ($ofs < $prev_ofs_plus_len) {
       printf STDERR "ERROR: overlapping region at end of %s:\n%s\n", $log_input1, $lastline;
   } elsif ($ofs > $prev_ofs_plus_len) {
       $gaplen = $ofs - $prev_ofs_plus_len;
       print LOG_OUTPUT &displayrange($prev_ofs_plus_len,$gaplen,'/') . "\n";
   }
   print LOG_OUTPUT "$lastline\n";
}

__END__

